class NoColumnsError(Exception):
    """
    Raised if no header are specified
    """
    pass


class NoTableNameGivenError(ValueError):
    """
    Raised if no header are specified
    """
    pass
